from enum import Enum
from typing import TypeVar, Optional, Union

from Bio.Seq import Seq
from inscripta.biocantor.exc import AlphabetError, NoSuchAncestorException, EmptySequenceFastaError
from inscripta.biocantor.location.strand import Strand
from inscripta.biocantor.parent import Parent, make_parent
from inscripta.biocantor.sequence.alphabet import (
    Alphabet,
    ALPHABET_TO_NUCLEOTIDE_COMPLEMENT,
)

Location = TypeVar("Location")
Sequence = TypeVar("Sequence")
ParentInputType = TypeVar("ParentInputType")


class Sequence:
    """A sequence with an alphabet"""

    sequence: Seq

    def __init__(
        self,
        data: str,
        alphabet: Alphabet,
        id: Optional[str] = None,
        type: Optional[str] = None,
        parent: Optional[ParentInputType] = None,
        validate_alphabet: bool = True,
    ):
        """
            Parameters
        ----------
            data
                The contents of the sequence
            alphabet
                Alphabet
            id
                Sequence name
            type
                Type of sequence
            parent
                Parent
            validate_alphabet
                Whether to validate this sequence against its alphabet
        """
        self.sequence = Seq(data)
        self.alphabet = alphabet
        self.id = id
        self.sequence_type = type
        self.parent = make_parent(parent) if parent else None
        self._len = len(self.sequence)
        if self.parent and self.parent.location and len(self.parent.location) != len(self):
            raise ValueError(
                "Sequence length ({}) does not equal parent location length ({})".format(
                    len(self), len(self.parent.location)
                )
            )
        if validate_alphabet:
            self._validate_alphabet()

    def __eq__(self, other):
        if type(other) is not Sequence:
            return False
        if self.id != other.id:
            return False
        if self.sequence_type != other.sequence_type:
            return False
        if self.alphabet != other.alphabet:
            return False
        if self.parent != other.parent:
            return False
        return self.sequence == other.sequence

    def __hash__(self):
        return hash((self.id, self.sequence_type, self.alphabet, self.parent, self.sequence))

    def __str__(self):
        """Returns the sequence data as a string"""
        return str(self.sequence)

    def __len__(self):
        return self._len

    def __getitem__(self, key: Union[int, slice]) -> Sequence:
        """Returns a slice of the current Sequence as a new Sequence object"""
        subseq = str(self.sequence[key])

        if self.parent is not None and self.parent.location is not None:
            if isinstance(key, slice):
                rel_start = key.start
                rel_end = key.stop
            else:
                rel_start = key
                rel_end = key + 1
            new_parent_location = self.parent.location.relative_interval_to_parent_location(
                relative_start=rel_start, relative_end=rel_end, relative_strand=Strand.PLUS
            )
            new_parent = self.parent.reset_location(new_parent_location)
        else:
            new_parent = self.parent

        return Sequence(subseq, self.alphabet, type=self.sequence_type, parent=new_parent)

    def __repr__(self):
        return "<{}>".format(self.summary())

    def summary(self) -> str:
        """Returns a short string summary of this Sequence"""
        if self.id:
            id = self.id
        else:
            if len(self) <= 20:
                id = "Sequence={}".format(str(self))
            else:
                id = "Sequence"
        return "{};\n  Alphabet={};\n  Length={};\n  Parent={};\n  Type={}".format(
            id, self.alphabet.name, len(self), repr(self.parent), self.sequence_type
        )

    def _validate_alphabet(self):
        """Raises AlphabetError if this Sequence does not conform to its alphabet"""
        Sequence.validate_alphabet(str(self), self.alphabet)

    @staticmethod
    def validate_alphabet(sequence, alphabet):
        if sequence.upper().strip(alphabet.value) != "":
            raise AlphabetError("Invalid sequence for alphabet {}".format(alphabet.name))

    @property
    def parent_id(self) -> Optional[str]:
        return self.parent.id if self.parent else None

    @property
    def is_empty(self) -> bool:
        """Is this a len 0 sequence?"""
        return self._len == 0

    @property
    def location_on_parent(self) -> Optional[Location]:
        """Location of this sequence relative to the parent"""
        return self.parent.location if self.parent else None

    @property
    def parent_strand(self) -> Optional[Strand]:
        """Strand of this sequence"""
        return self.parent.strand if self.parent else None

    @property
    def parent_type(self) -> Optional[str]:
        return self.parent.sequence_type if self.parent else None

    def reverse_complement(self, new_id: str = None, new_type: str = None) -> "Sequence":
        """Returns a new Sequence corresponding to the reverse complement of this Sequence.
        Location on parent, if it exists, is converted appropriately.

        Parameters
        ----------
        new_id
            ID for the returned Sequence. If no value is provided, None is used.
        new_type
            Sequence type for the returned Sequence. If no value is provided, None is used.
        """
        if not self.alphabet.is_nucleotide_alphabet():
            raise AlphabetError("Cannot reverse complement sequence with alphabet {}".format(self.alphabet))
        location = self.location_on_parent.reverse_strand() if self.location_on_parent else None
        strand = self.parent_strand.reverse() if self.parent_strand else None
        rc_map = ALPHABET_TO_NUCLEOTIDE_COMPLEMENT[self.alphabet]
        try:
            seq_data = "".join([rc_map[c] for c in str(self)[::-1]])
        except KeyError as e:
            raise AlphabetError("Character {} not found for alphabet {}".format(str(e), self.alphabet))
        rc_parent = Parent(strand=strand, location=location) if strand or location else None
        return Sequence(
            seq_data,
            self.alphabet,
            id=new_id,
            type=new_type,
            parent=rc_parent,
            validate_alphabet=False,
        )

    def append(self, other: Sequence, new_id: Optional[str] = None, data_only: bool = False) -> Sequence:
        """Returns a new Sequence consisting of other sequence appended to the end of this Sequence

        Parameters
        ----------
        other
            Other sequence
        new_id
            ID for the returned sequence
        data_only
            Only combine sequence data, ignoring and removing metadata such as sequence type and parent
            info. If False, all metadata for both sequences must be consistent, and the returned sequence
            will have appropriate combined metadata.
        """
        if self.alphabet != other.alphabet:
            raise ValueError("Sequences must have same alphabet: {} != {}".format(self.alphabet, other.alphabet))
        new_seq_data = "{}{}".format(str(self), str(other))
        if data_only:
            return Sequence(new_seq_data, self.alphabet, id=new_id)
        if self.sequence_type != other.sequence_type:
            raise ValueError("Sequences must have same type: {} != {}".format(self.sequence_type, other.sequence_type))
        if self.parent:
            if not self.parent.equals_except_location(other.parent):
                raise ValueError(
                    "Sequences must have same parent (except location on parent):\n{}\n  !=\n{}".format(
                        repr(self.parent), repr(other.parent)
                    )
                )
            if self.parent.strand == Strand.UNSTRANDED or self.parent.strand != other.parent.strand:
                raise ValueError("Invalid strands on parent: {} != {}".format(self.parent.strand, other.parent.strand))
            if self.parent.location and other.parent.location:
                if self.parent.strand == Strand.PLUS and self.parent.location.end > other.parent.location.start:
                    raise ValueError("Sequence on plus strand of parent must be to the left of appended sequence")
                if self.parent.strand == Strand.MINUS and self.parent.location.start < other.parent.location.end:
                    raise ValueError("Sequence on minus strand of parent must be to the right of appended sequence")
                new_location = self.parent.location.union(other.parent.location)
            else:
                new_location = None
            new_parent = self.parent.strip_location_info().reset_location(new_location)
        else:
            new_parent = None
        return Sequence(
            new_seq_data,
            self.alphabet,
            id=new_id,
            type=self.sequence_type,
            parent=new_parent,
            validate_alphabet=False,
        )

    def first_ancestor_of_type(self, sequence_type: str, include_self: bool = True) -> Parent:
        """Returns the Parent object representing the closest ancestor (parent, parent of parent, etc.)
        of this sequence which has the given sequence type. If include_self is True and this sequence has the
        given type, returns a new Parent object representing this sequence. Raises NoSuchAncestorException
        if no ancestor with the given type exists.

        Parameters
        ----------
        sequence_type:
            Sequence type
        include_self:
            Include this sequence as a candidate
        """
        if include_self and self.sequence_type == sequence_type:
            return Parent(sequence=self)
        if self.parent:
            return self.parent.first_ancestor_of_type(sequence_type, True)
        raise NoSuchAncestorException

    def has_ancestor_of_type(self, sequence_type: str, include_self: bool = True) -> bool:
        """Returns True if some ancestor (parent, parent of parent, etc.) of this sequence which has the given sequence
        type, or False otherwise. If include_self is True and this sequence has the given type, returns True.

        Parameters
        ----------
        sequence_type:
            Sequence type
        include_self:
            Include this sequence as a candidate
        """
        if include_self and self.sequence_type == sequence_type:
            return True
        if self.parent:
            return self.parent.has_ancestor_of_type(sequence_type, include_self=True)
        return False

    def to_fasta(self, num_chars: Optional[int] = 60) -> str:
        """Returns a FASTA-formatted string for this sequence. These are line-broken every num_chars.

        Parameters
        ----------
        num_chars:
            Number of characters per line. Defaults to 60, which is the same as BioPython.
        """
        if self.is_empty:
            raise EmptySequenceFastaError("Cannot write FASTA for empty Sequence")

        r = [f">{self.id}"]
        for i in range(0, self._len, num_chars):
            r.append(str(self)[i : i + num_chars])
        return "\n".join(r)



from enum import Enum


class Alphabet(Enum):
    NT_STRICT = "ACGT"
    NT_EXTENDED = "ATUCGNWSMKRYBDHV"
    NT_STRICT_GAPPED = "ACGT-"
    NT_EXTENDED_GAPPED = "ATUCGNWSMKRYBDHV-"
    AA = "GALMFWKQESPVICYHRNDT*"
    GENERIC = "ABCDEFGHIJKLMNOPQRSTUVWXYZ-"

    def is_nucleotide_alphabet(self):
        if self in [
            Alphabet.NT_STRICT,
            Alphabet.NT_EXTENDED,
            Alphabet.NT_STRICT_GAPPED,
            Alphabet.NT_EXTENDED_GAPPED,
        ]:
            return True
        if self in [Alphabet.AA, Alphabet.GENERIC]:
            return False
        raise NotImplementedError("Not implemented for alphabet {}".format(self))


ALPHABET_TO_NUCLEOTIDE_COMPLEMENT = {
    Alphabet.NT_STRICT: {"A": "T", "a": "t", "C": "G", "c": "g", "G": "C", "g": "c", "T": "A", "t": "a"},
    Alphabet.NT_EXTENDED: {
        "A": "T",
        "a": "t",
        "T": "A",
        "t": "a",
        "U": "A",
        "u": "a",
        "G": "C",
        "g": "c",
        "C": "G",
        "c": "g",
        "Y": "R",
        "y": "r",
        "R": "Y",
        "r": "y",
        "S": "S",
        "s": "s",
        "W": "W",
        "w": "w",
        "K": "M",
        "k": "m",
        "M": "K",
        "m": "k",
        "B": "V",
        "b": "v",
        "D": "H",
        "d": "h",
        "H": "D",
        "h": "d",
        "V": "B",
        "v": "b",
        "N": "N",
        "n": "n",
    },
    Alphabet.NT_STRICT_GAPPED: {
        "A": "T",
        "a": "t",
        "C": "G",
        "c": "g",
        "G": "C",
        "g": "c",
        "T": "A",
        "t": "a",
        "-": "-",
    },
    Alphabet.NT_EXTENDED_GAPPED: {
        "A": "T",
        "a": "t",
        "T": "A",
        "t": "a",
        "U": "A",
        "u": "a",
        "G": "C",
        "g": "c",
        "C": "G",
        "c": "g",
        "Y": "R",
        "y": "r",
        "R": "Y",
        "r": "y",
        "S": "S",
        "s": "s",
        "W": "W",
        "w": "w",
        "K": "M",
        "k": "m",
        "M": "K",
        "m": "k",
        "B": "V",
        "b": "v",
        "D": "H",
        "d": "h",
        "H": "D",
        "h": "d",
        "V": "B",
        "v": "b",
        "N": "N",
        "n": "n",
        "-": "-",
    },
}



from enum import Enum
from typing import List

from inscripta.biocantor.constants import gencode, aacodons


class Codon(Enum):

    ATA = "ATA"
    ATC = "ATC"
    ATT = "ATT"
    ATG = "ATG"
    ACA = "ACA"
    ACC = "ACC"
    ACG = "ACG"
    ACT = "ACT"
    AAC = "AAC"
    AAT = "AAT"
    AAA = "AAA"
    AAG = "AAG"
    AGC = "AGC"
    AGT = "AGT"
    AGA = "AGA"
    AGG = "AGG"
    CTA = "CTA"
    CTC = "CTC"
    CTG = "CTG"
    CTT = "CTT"
    CCA = "CCA"
    CCC = "CCC"
    CCG = "CCG"
    CCT = "CCT"
    CAC = "CAC"
    CAT = "CAT"
    CAA = "CAA"
    CAG = "CAG"
    CGA = "CGA"
    CGC = "CGC"
    CGG = "CGG"
    CGT = "CGT"
    GTA = "GTA"
    GTC = "GTC"
    GTG = "GTG"
    GTT = "GTT"
    GCA = "GCA"
    GCC = "GCC"
    GCG = "GCG"
    GCT = "GCT"
    GAC = "GAC"
    GAT = "GAT"
    GAA = "GAA"
    GAG = "GAG"
    GGA = "GGA"
    GGC = "GGC"
    GGG = "GGG"
    GGT = "GGT"
    TCA = "TCA"
    TCC = "TCC"
    TCG = "TCG"
    TCT = "TCT"
    TTC = "TTC"
    TTT = "TTT"
    TTA = "TTA"
    TTG = "TTG"
    TAC = "TAC"
    TAT = "TAT"
    TAA = "TAA"
    TAG = "TAG"
    TGC = "TGC"
    TGT = "TGT"
    TGA = "TGA"
    TGG = "TGG"

    def translate(self) -> str:
        """Returns string symbol of translated amino acid"""
        return gencode[self.value]

    def synonymous_codons(self, include_self=False) -> List["Codon"]:
        """Returns list of synonymous codons

        Parameters
        ----------
        include_self
            Include this codon in returned list
        """
        codons = [Codon(codon_str) for codon_str in aacodons[self.translate()]]
        if not include_self:
            codons.remove(self)
        return codons

    @property
    def is_stop_codon(self) -> bool:
        return self.value in aacodons["*"]



from functools import reduce
from typing import TypeVar, Optional

import inscripta.biocantor
from inscripta.biocantor.exc import NoSuchAncestorException
from inscripta.biocantor.location.strand import Strand
from inscripta.biocantor.util.object_validation import ObjectValidation

Parent = TypeVar("Parent")
ParentInputType = TypeVar("ParentInputType")


class Parent:
    """
    Holds information about a parent of some object. Typically the child object should hold
    a reference to this parent.
    """

    def __init__(
        self,
        *,
        id: Optional[str] = None,
        sequence_type: Optional[str] = None,
        strand: Optional[Strand] = None,
        location: Optional = None,
        sequence: Optional = None,
        parent: Optional[ParentInputType] = None,
    ):
        """
        Parameters
        ----------
        id
            Parent ID
        sequence_type: string
            Parent type
        strand: Strand
            Strand of child object with respect to this parent
        location: Location
            Location of child object with respect to this parent
        sequence: Sequence
            Sequence of this parent
        parent: Parent
            Parent of this parent
        """

        location_parent_id = location.parent_id if location else None
        sequence_id = sequence.id if sequence else None
        non_null_ids = set([x for x in [id, location_parent_id, sequence_id] if x is not None])
        if len(non_null_ids) > 1:
            raise ValueError(
                "ID, location parent ID, and sequence ID do not match: {}, {}, {}".format(
                    id, location_parent_id, sequence_id
                )
            )

        location_parent_type = location.parent_type if location else None
        sequence_seqtype = sequence.sequence_type if sequence else None
        non_null_types = set([x for x in [sequence_type, location_parent_type, sequence_seqtype] if x is not None])
        if len(non_null_types) > 1:
            raise ValueError(
                "Sequence type, location parent type, and sequence do not match: {}, {}, {}".format(
                    sequence_type, location_parent_type, sequence_seqtype
                )
            )

        if location:
            if strand and location.strand and strand != location.strand:
                raise ValueError("Strand does not match location: {} != {}".format(strand, location.strand))
            if sequence and location.end > len(sequence):
                raise ValueError(
                    "Location end ({}) is greater than sequence length ({})".format(location.end, len(sequence))
                )

        parent_obj = inscripta.biocantor.parent.make_parent(parent) if parent else None
        if sequence and parent_obj and parent_obj.sequence and len(sequence) > len(parent_obj.sequence):
            raise ValueError(
                "Parent ({}) is longer than parent of parent ({})".format(len(sequence), len(parent_obj.sequence))
            )

        if sequence and sequence.parent:
            if parent_obj:
                ObjectValidation.require_parents_equal_except_location(parent_obj, sequence.parent)
                self.parent = parent_obj
            else:
                self.parent = sequence.parent
        else:
            self.parent = parent_obj

        self._id = non_null_ids.pop() if non_null_ids else None
        self._sequence_type = non_null_types.pop() if non_null_types else None
        self._strand = strand
        self.location = location
        self.sequence = sequence

    def __eq__(self, other):
        if not self.equals_except_location(other):
            return False
        return self.location == other.location and self.strand == other.strand

    def equals_except_location(self, other):
        if type(other) is not Parent:
            return False
        if self.id != other.id:
            return False
        if self.sequence_type != other.sequence_type:
            return False
        if self.sequence != other.sequence:
            return False
        if self.parent and other.parent and self.parent != other.parent:
            return False
        return True

    def __hash__(self):
        return hash(
            (
                self.id,
                self.sequence_type,
                self.strand,
                self.location,
                self.sequence,
                self.parent,
            )
        )

    def __repr__(self):
        return "<Parent: id={}, type={}, strand={}, location={}, sequence={}, parent={}>".format(
            self.id,
            self.sequence_type,
            self.strand,
            repr(self.location),
            repr(self.sequence),
            repr(self.parent),
        )

    @property
    def id(self):
        if self._id:
            return self._id
        if self.sequence:
            return self.sequence.id
        return None

    @property
    def sequence_type(self):
        if self._sequence_type:
            return self._sequence_type
        if self.sequence:
            return self.sequence.sequence_type
        return None

    @property
    def strand(self):
        if self._strand:
            return self._strand
        if self.location:
            return self.location.strand
        return None

    def strip_location_info(self) -> Parent:
        """Returns a new Parent object representing this Parent with information about child
        location removed"""
        return Parent(
            id=self.id,
            sequence_type=self.sequence_type,
            sequence=self.sequence,
            parent=self.parent,
        )

    def first_ancestor_of_type(self, sequence_type, include_self: bool = True) -> Parent:
        """Returns the Parent object representing the closest ancestor (parent, parent of parent, etc.)
        of this Parent which has the given sequence type. If include_self is True and this Parent
        has the given type, returns this object. Raises NoSuchAncestorException if no ancestor with the given
        type exists.

        Parameters
        ----------
        sequence_type: str
            Sequence type
        include_self:
            Include this sequence as a candidate
        """
        if include_self and self.sequence_type == sequence_type:
            return self
        if self.parent:
            return self.parent.first_ancestor_of_type(sequence_type)
        raise NoSuchAncestorException

    def has_ancestor_of_type(self, sequence_type, include_self: bool = True) -> bool:
        """Returns True if some ancestor (parent, parent of parent, etc.) of this Parent has the given sequence type,
        or False otherwise. If include_self is True and this Parent has the given type, returns True.

        Parameters
        ----------
        sequence_type: str
            Sequence type
        include_self:
            Include this sequence as a candidate
        """
        if include_self and self.sequence_type == sequence_type:
            return True
        if self.parent:
            return self.parent.has_ancestor_of_type(sequence_type, include_self=True)
        return False

    def lift_child_location_to_parent(self):
        """Lifts location of child object on this parent to the parent of this parent.
        Raises ValueError if any required data is missing (child location or location of this parent
        on its parent).

        Returns
        -------
        Location
            Child object location lifted to the parent of this parent
        """
        ObjectValidation.require_parent_has_location(self)
        ObjectValidation.require_parent_has_parent_with_location(self)
        lifted_blocks = [
            self.parent.location.relative_interval_to_parent_location(block.start, block.end, block.strand)
            for block in self.location.blocks
        ]
        lifted_blocks_union = reduce(lambda location1, location2: location1.union(location2), lifted_blocks)
        location_with_parent = lifted_blocks_union.reset_parent(self.parent.strip_location_info())
        return location_with_parent

    def reset_location(self, location) -> "Parent":
        """Returns a new Parent object with child location set to the given location"""
        strand = location.strand if location else None
        return Parent(
            id=self.id,
            sequence_type=self.sequence_type,
            strand=strand,
            location=location,
            sequence=self.sequence,
            parent=self.parent,
        )

    def has_ancestor_sequence(self, sequence, include_self: bool = True) -> bool:
        """Returns True iff this Parent has some ancestor (parent, parent of parent, etc.) whose sequence
        attribute is equal to the given sequence. If include_self is True and this Parent has sequence
        attribute equal to the given sequence, returns True."""
        if include_self and self.sequence and self.sequence == sequence:
            return True
        if not self.parent:
            return False
        return self.parent.has_ancestor_sequence(sequence, include_self=True)



from enum import Enum
from functools import total_ordering
from inscripta.biocantor.exc import UnsupportedOperationException, InvalidStrandException


@total_ordering
class Strand(Enum):
    PLUS = 1
    MINUS = -1
    UNSTRANDED = 0

    def __str__(self):
        return str(self.to_symbol())

    @staticmethod
    def from_symbol(value: str):
        """Converts string representation of a strand to a Strand"""
        if value == "+":
            return Strand.PLUS
        if value == "-":
            return Strand.MINUS
        raise ValueError("{} is not a valid string representation of a strand".format(value))

    def to_symbol(self) -> str:
        if self == Strand.PLUS:
            return "+"
        if self == Strand.MINUS:
            return "-"
        return "."

    @staticmethod
    def from_int(value: int):
        """Converts integer representation of a strand to a Strand"""
        return Strand(value)  # Raises ValueError for invalid int

    @staticmethod
    def _order():
        return {Strand.PLUS: 1, Strand.MINUS: 2, Strand.UNSTRANDED: 3}

    def __lt__(self, other):
        if not type(other) is Strand:
            raise ValueError("Cannot compare {} to {}".format(type(self).__name__, type(other).__name__))
        order = Strand._order()
        return order[self] < order[other]

    def reverse(self):
        """Returns the opposite of this Strand"""
        if self == Strand.PLUS:
            return Strand.MINUS
        if self == Strand.MINUS:
            return Strand.PLUS
        if self == Strand.UNSTRANDED:
            return Strand.UNSTRANDED
        raise UnsupportedOperationException("Not implemented for {}".format(self))

    def relative_to(self, other: "Strand") -> "Strand":
        """Returns the orientation of this strand relative to the given other strand.
        Note: this operator is commutative.
        """
        if self == Strand.UNSTRANDED or other == Strand.UNSTRANDED:
            return Strand.UNSTRANDED
        if self == other:
            return Strand.PLUS
        return Strand.MINUS

    def assert_directional(self):
        """Raises InvalidStrandException if this Strand does not have a defined direction (plus or minus)"""
        if self not in [Strand.PLUS, Strand.MINUS]:
            raise InvalidStrandException("Strand {} does not have a defined direction".format(self))



from inscripta.biocantor.exc import LocationException, LocationOverlapException


class ObjectValidation:
    @staticmethod
    def require_location_nonempty(location):
        if len(location) < 1:
            raise LocationException("Location must have positive length:\n{}".format(repr(location)))

    @staticmethod
    def require_location_has_parent(location):
        if not location.parent:
            raise LocationException("Location must have non-null parent attribute:\n{}".format(repr(location)))

    @staticmethod
    def require_location_has_parent_with_sequence(location):
        ObjectValidation.require_location_has_parent(location)
        if not location.parent.sequence:
            raise ValueError(
                "Parent of location must have non-null sequence attribute:\n{}".format(repr(location.parent))
            )

    @staticmethod
    def require_parent_has_location(parent):
        if not parent.location:
            raise ValueError("Parent must have non-null location attribute:\n{}".format(repr(parent)))

    @staticmethod
    def require_parent_has_parent(parent):
        if not parent.parent:
            raise ValueError("Parent must have non-null parent attribute:\n{}".format(repr(parent)))

    @staticmethod
    def require_parent_has_parent_with_location(parent):
        ObjectValidation.require_parent_has_parent(parent)
        if not parent.parent.location:
            raise ValueError(
                "Parent must have parent attribute with non-null location attribute:\n{}".format(repr(parent))
            )

    @staticmethod
    def require_parents_equal_except_location(parent1, parent2):
        if not parent1.equals_except_location(parent2):
            raise ValueError("Parents must be equal except location info:\n{}\n  !=\n{}".format(parent1, parent2))

    @staticmethod
    def require_locations_have_same_nonempty_parent(location1, location2):
        ObjectValidation.require_location_has_parent(location1)
        ObjectValidation.require_location_has_parent(location2)
        ObjectValidation.require_parents_equal_except_location(location1.parent, location2.parent)

    @staticmethod
    def require_locations_overlap(location1, location2, match_strand: bool = False):
        if not location1.has_overlap(location2, match_strand=match_strand):
            raise LocationOverlapException("Locations must overlap:\n{}\n{}".format(repr(location1), repr(location2)))

    @staticmethod
    def require_locations_do_not_overlap(location1, location2, match_strand: bool = False):
        if location1.has_overlap(location2, match_strand=match_strand):
            raise LocationOverlapException(
                "Locations must not overlap:\n{}\n{}".format(repr(location1), repr(location2))
            )

    @staticmethod
    def require_object_has_type(obj, required_type):
        if type(obj) is not required_type:
            raise TypeError("Object must have type {}".format(required_type))



class InvalidStrandException(Exception):
    pass


class InvalidPositionException(Exception):
    pass


class UnsupportedOperationException(TypeError):
    pass


class LocationException(ValueError):
    pass


class LocationOverlapException(LocationException):
    pass


class EmptyLocationException(LocationException):
    pass


class AlphabetError(ValueError):
    pass


class NoSuchAncestorException(Exception):
    pass


class ValidationError(Exception):
    pass


class InvalidCDSIntervalError(Exception):
    pass


class EmptySequenceFastaError(Exception):
    pass



from enum import Enum
from typing import Iterator, List, Union, Optional, AnyStr

from inscripta.biocantor.exc import ValidationError, EmptyLocationException
from inscripta.biocantor.gene.codon import Codon
from inscripta.biocantor.location.location import Location, Strand
from inscripta.biocantor.location.location_impl import SingleInterval
from inscripta.biocantor.sequence.alphabet import Alphabet
from inscripta.biocantor.sequence import Sequence


class CDSPhase(Enum):
    """
    It is important not to confuse Phase with Frame. From the GFF3 specification:

    The phase is one of the integers 0, 1, or 2, indicating the number of bases forward from the start of the
    current CDS feature the next codon begins. A phase of "0" indicates that a codon begins on the first nucleotide
    of the CDS feature (i.e. 0 bases forward), a phase of "1" indicates that the codon begins at the second nucleotide
    of this CDS feature and a phase of "2" indicates that the codon begins at the third nucleotide of this region.

    """

    NONE = -1
    ZERO = 0
    ONE = 1
    TWO = 2

    @staticmethod
    def from_int(value: int) -> "CDSPhase":
        return CDSPhase(value)  # Raises ValueError for invalid int

    def to_frame(self) -> "CDSFrame":
        """
        https://github.com/ucscGenomeBrowser/kent/blob/022eb4f62a0af16526ca1bebcd9e68bd456265dc/src/inc/gff3.h#L281-L293
        """
        mapping = {0: 0, 2: 1, 1: 2, -1: -1}
        return CDSFrame(mapping[self.value])

    def to_gff(self) -> AnyStr:
        """In GFF format, Phase is represented with a period for NONE"""
        if self == CDSPhase.NONE:
            return "."
        return str(self.value)


class CDSFrame(Enum):
    """
    From the GFF3 specification:

    Frame is generally calculated as a value for a given base relative to the start of the complete
    open reading frame (ORF) or the codon (e.g. modulo 3) while CDS phase describes the start of the next codon
    relative to a given CDS feature.

    Frame is easier to work with computationally because for any given position it can be calculated by the distance
    from transcription start % 3. Care must be taken for strand -- if the CDS is on the minus strand, then this
    calculation instead becomes distance from transcription stop % 3.
    """

    NONE = -1
    ZERO = 0
    ONE = 1
    TWO = 2

    @staticmethod
    def from_int(value: int) -> "CDSFrame":
        return CDSFrame(value)

    def shift(self, shift: int) -> "CDSFrame":
        if self is CDSFrame.NONE:
            return self
        if shift > 0:
            return CDSFrame.from_int((self.value + shift) % 3)
        else:
            return CDSFrame.from_int((self.value - (shift - ((-shift) % 3))) % 3)

    def to_phase(self) -> "CDSPhase":
        """Converts frame to phase"""
        mapping = {0: 0, 1: 2, 2: 1, -1: -1}
        return CDSPhase(mapping[self.value])


class CDSInterval:
    """
    An wrapper for a Location that gives it a concept of Frames.
    Frame must be recorded individually on *every interval*.

    This is necessary to be able to encode programmed frame shifts and indel variation that
    would break frame, but that are likely just errors in assembly or annotation.
    """

    def __init__(self, location: Location, frames: List[Union[CDSFrame, CDSPhase]]):
        self.location = location
        if len(frames) != location.num_blocks:
            raise ValidationError("Number of frame entries must match number of exons")
        # internally we work with Frame, but support Phase
        # this will make parsing GFF3 easier
        for i, frame in enumerate(frames):
            if isinstance(frame, CDSPhase):
                frame = frame.to_frame()
                frames[i] = frame
        self.frames = frames

    def __str__(self):
        frame_str = ", ".join([str(p) for p in self.frames])
        return f"CDS(({self.location}), ({frame_str})"

    def __repr__(self):
        return "<{}>".format(str(self))

    def __eq__(self, other):
        if type(other) is not CDSInterval:
            return False
        elif self.frames != other.frames:
            return False
        return self.location == other.location

    def __len__(self) -> int:
        return len(self.location)

    @property
    def has_valid_stop(self) -> bool:
        """Does this CDS have a valid stop? Requires a sequence be associated."""
        seq = self.extract_sequence()
        c = Codon(seq[-3:].sequence)
        return c.is_stop_codon

    @property
    def strand(self) -> Strand:
        """Pass up the Strand of this CDS's Location"""
        return self.location.strand

    @property
    def start(self) -> int:
        """Pass up the start of this CDS's Location"""
        return self.location.start

    @property
    def end(self) -> int:
        """Pass up the end of this CDS's Location"""
        return self.location.end

    def intersect(self, location: Location) -> "CDSInterval":
        """
        Returns a new CDS representing the intersection of this CDS's location with the other location.
        Strand of the other location is ignored; returned CDS is on the same strand as this CDS.
        """
        intersection = self.location.intersection(location, match_strand=False)

        if intersection.is_empty:
            raise EmptyLocationException("Can't intersect disjoint intervals")

        frames = []
        if self.location.strand == Strand.PLUS or self.location.strand == Strand.UNSTRANDED:
            frame_shift = self.location.start - intersection.start
        else:
            frame_shift = intersection.end - self.location.end

        # adjust original frames by the new frame shift value
        for block, frame in zip(self.location.blocks, self.frames):
            if block.has_overlap(intersection):
                frames.append(frame.shift(frame_shift))

        return CDSInterval(intersection, frames)

    def frame_iter(self) -> Iterator[CDSFrame]:
        """Iterate over frames taking strand into account"""
        if self.location.strand == Strand.PLUS or self.location.strand == Strand.UNSTRANDED:
            yield from self.frames
        else:
            yield from reversed(self.frames)

    def exon_iter(self) -> Iterator[SingleInterval]:
        """Iterate over exons in transcription direction"""
        if self.location.strand == Strand.PLUS or self.location.strand == Strand.UNSTRANDED:
            yield from self.location.blocks
        else:
            yield from reversed(self.location.blocks)

    def extract_sequence(self) -> Sequence:
        """
        Returns a continuous CDS sequence that is in frame and always a multiple of 3.

        Any leading or trailing bases that are annotated as CDS but cannot form a full codon
        are removed. Additionally, any internal codons that are incomplete are removed.

        Incomplete internal codons are determined by comparing the CDSFrame of each exon
        as annotated, to the expected value of the CDSFrame. This allows for an annotation
        to model things like programmed frameshifts and indels that may be assembly errors.
        """
        seq = []
        # keeps track of what we expect the next frame to be
        next_frame = CDSFrame(CDSFrame.ZERO)
        for exon, frame in zip(self.exon_iter(), self.frame_iter()):
            s = list(exon.extract_sequence())
            if next_frame != frame:
                s = s[frame.value :]
                # remove trailing codon from previous sequence
                shift = len(seq) % 3
                if shift > 0:
                    seq = seq[:-shift]
                # we are now inherently in frame
                next_frame = CDSFrame(CDSFrame.ZERO)
            seq.extend(s)
            # this is what we expect the next frame to be, if no frameshift occurred
            next_frame = next_frame.shift(len(s))
        # we may not have a complete 3' end
        extra = len(seq) % 3
        if extra > 0:
            seq = "".join(str(s) for s in seq[:-extra])
        else:
            seq = "".join(str(s) for s in seq)
        assert len(seq) % 3 == 0
        return Sequence(seq, Alphabet.NT_EXTENDED)

    def scan_codons(self, truncate_at_in_frame_stop: Optional[bool] = False) -> Iterator[Codon]:
        """
        Iterator along codons. If truncate_at_in_frame_stop is True,
        this will stop iteration at the first in-frame stop.
        """
        seq = self.extract_sequence()
        for i in range(0, len(seq), 3):
            c = Codon(str(seq[i : i + 3]).upper())
            yield c
            if truncate_at_in_frame_stop and c.is_stop_codon:
                break

    def translate(self, truncate_at_in_frame_stop: Optional[bool] = False) -> Sequence:
        """
        Returns amino acid sequence of this CDS. If truncate_at_in_frame_stop is True,
        this will stop at the first in-frame stop
        """
        aa_seq_str = "".join([codon.translate() for codon in self.scan_codons(truncate_at_in_frame_stop)])
        return Sequence(aa_seq_str, Alphabet.AA)



from inscripta.biocantor.sequence.sequence import Sequence
from inscripta.biocantor.parent import make_parent, Parent


@make_parent.register(Sequence)
def _(obj) -> Parent:
    return Parent(sequence=obj)



from inscripta.biocantor.location.location import Location
from inscripta.biocantor.location.strand import Strand
from inscripta.biocantor.parent import make_parent, Parent


@make_parent.register(Location)
def _(obj) -> Parent:
    return Parent(location=obj)


@make_parent.register(Strand)
def _(obj) -> Parent:
    return Parent(strand=obj)



from functools import singledispatch

from inscripta.biocantor.parent.parent import Parent


@singledispatch
def make_parent(obj) -> Parent:
    raise TypeError("{} not supported".format(type(obj)))


@make_parent.register(str)
def _(obj) -> Parent:
    return Parent(id=obj)


@make_parent.register(Parent)
def _(obj) -> Parent:
    return obj



from typing import Union

from inscripta.biocantor.location.strand import Strand
from inscripta.biocantor.parent import Parent
from inscripta.biocantor.sequence import Sequence
from inscripta.biocantor.location.location import Location


ParentInputType = Union[Sequence, str, Location, Strand, Parent]



from functools import total_ordering, reduce
from typing import Optional, List, Iterator

from Bio.SeqFeature import FeatureLocation, CompoundLocation
from inscripta.biocantor.util.types import ParentInputType
from inscripta.biocantor.exc import (
    InvalidStrandException,
    InvalidPositionException,
    UnsupportedOperationException,
    EmptyLocationException,
)
from inscripta.biocantor.location.distance import DistanceType
from inscripta.biocantor.location.location import Location
from inscripta.biocantor.location.strand import Strand
from inscripta.biocantor.parent import Parent, make_parent
from inscripta.biocantor.sequence import Sequence
from inscripta.biocantor.util.object_validation import ObjectValidation


@total_ordering
class SingleInterval(Location):
    """A single contiguous interval within a sequence"""

    def __init__(
        self,
        start: int,
        end: int,
        strand: Strand,
        parent: Optional[ParentInputType] = None,
    ):
        """
        Parameters
        ----------
        start
            0-based start position of this Location on parent
        end
            0-based exclusive end position of this Location on parent
        strand
            Strand of this Location on parent
        parent
            Parent containing this Location. If parent is not provided including a Sequence,
            some methods requiring the sequence cannot be called. Additionally, some position validation
            cannot be performed. If parent has location attribute, location is ignored and set to this location.
        """
        if not 0 <= start <= end:
            raise InvalidPositionException(f"Positions must satisfy 0 <= start <= end. Start: {start}, end: {end}")
        parent_obj = make_parent(parent) if parent else None
        if parent_obj and parent_obj.sequence:
            if not start < len(parent_obj.sequence):
                raise InvalidPositionException(
                    f"Start position ({start}) must be < parent length ({len(parent_obj.sequence)})"
                )
            if not end <= len(parent_obj.sequence):
                raise InvalidPositionException(
                    f"End position ({end}) must be <= parent length ({len(parent_obj.sequence)})"
                )

        self._start = start
        self._end = end
        self._strand = strand
        self._parent = parent_obj.reset_location(SingleInterval(start, end, strand)) if parent_obj else None
        self._sequence = None

    def __str__(self):
        return f"{self.start}-{self.end}:{self.strand}"

    def __repr__(self):
        data = f"{repr(self.parent)}:{str(self)}" if self.parent else str(self)
        return f"<SingleInterval {data}>"

    def __len__(self):
        return self._end - self._start

    @property
    def parent(self) -> Parent:
        return self._parent

    @property
    def strand(self) -> Strand:
        return self._strand

    @property
    def start(self) -> int:
        return self._start

    @property
    def end(self) -> int:
        return self._end

    @property
    def is_contiguous(self) -> bool:
        return True

    @property
    def is_empty(self) -> bool:
        return self == EmptyLocation()

    @property
    def blocks(self) -> List[Location]:
        return [self]

    def scan_blocks(self) -> Iterator[Location]:
        yield self

    @property
    def num_blocks(self) -> int:
        return 1

    @property
    def is_overlapping(self) -> bool:
        """SingleInterval is by definition always non-overlapping"""
        return False

    def __eq__(self, other):
        if type(other) is not SingleInterval:
            return False
        return (
            self.start == other.start
            and self.end == other.end
            and self.strand == other.strand
            and self.parent == other.parent
        )

    def optimize_blocks(self) -> Location:
        if len(self) == 0:
            return EmptyLocation()
        return self

    def gap_list(self) -> List["Location"]:
        return []

    def gaps_location(self) -> "Location":
        return EmptyLocation()

    def __hash__(self):
        return hash((self.start, self.end, self.strand, self.parent))

    def __lt__(self, other: Location):
        self_parent_id = self.parent.id if self.parent is not None else ""
        other_parent_id = other.parent.id if other.parent is not None else ""
        if self_parent_id != other_parent_id:
            return self_parent_id < other_parent_id
        if self.start != other.start:
            return self.start < other.start
        if self.end != other.end:
            return self.end < other.end
        return self.strand < other.strand

    def extract_sequence(self) -> Sequence:
        if self._sequence is None:
            ObjectValidation.require_location_has_parent_with_sequence(self)
            seq_plus_strand = str(self.parent.sequence)[self.start : self.end]
            if self.strand == Strand.PLUS:
                self._sequence = Sequence(
                    seq_plus_strand,
                    self.parent.sequence.alphabet,
                    validate_alphabet=False,
                )
            elif self.strand == Strand.MINUS:
                self._sequence = Sequence(
                    seq_plus_strand,
                    self.parent.sequence.alphabet,
                    validate_alphabet=False,
                ).reverse_complement()
            else:
                raise InvalidStrandException(f"Invalid strand: {self.strand}")
        return self._sequence

    def parent_to_relative_pos(self, parent_pos: int) -> int:
        if parent_pos < self.start or parent_pos >= self.end:
            raise InvalidPositionException(
                f"Requested parent position {parent_pos} lies outside this location ({self.start}-{self.end})"
            )
        if self.strand == Strand.PLUS:
            return parent_pos - self.start
        if self.strand == Strand.MINUS:
            return self.end - parent_pos - 1
        raise InvalidStrandException(f"Invalid strand: {self.strand}")

    def relative_to_parent_pos(self, relative_pos: int) -> int:
        if relative_pos < 0 or relative_pos >= len(self):
            raise ValueError(
                f"Requested position ({relative_pos}) is not between 0 and length of this location ({len(self)})"
            )
        if self.strand == Strand.PLUS:
            return self.start + relative_pos
        if self.strand == Strand.MINUS:
            return self.end - relative_pos - 1
        raise InvalidStrandException(f"Invalid strand: {self.strand}")

    def relative_interval_to_parent_location(
        self, relative_start: int, relative_end: int, relative_strand: Strand
    ) -> Location:
        if not 0 <= relative_start <= relative_end <= len(self):
            raise ValueError(
                f"Invalid relative interval for interval of length {len(self)}: {relative_start}-{relative_end}"
            )
        if self.strand == Strand.PLUS:
            parent_start = self.start + relative_start
            parent_end = self.start + relative_end
        elif self.strand == Strand.MINUS:
            parent_start = self.end - relative_end
            parent_end = self.end - relative_start
        else:
            raise InvalidStrandException(f"Location strand must be {Strand.PLUS} or {Strand.MINUS}")
        new_parent = self.parent.strip_location_info() if self.parent else None
        return SingleInterval(
            parent_start,
            parent_end,
            self.strand.relative_to(relative_strand),
            parent=new_parent,
        )

    def has_overlap(self, other: Location, match_strand: bool = False) -> bool:
        # Todo: Should this always be 'False'?  If our intervals are on different chromosomes, they can't overlap
        #   (chr1:100-200 doesn't overlap chr2:150-250) but just because they have different parents doesn't mean
        #   they can't overlap...
        if self.parent_id != other.parent_id:
            return False
        if self.parent or other.parent:
            if not (self.parent and other.parent):
                return False
            if not self.parent.equals_except_location(other.parent):
                return False
        if match_strand and self.strand != other.strand:
            return False
        if type(other) is SingleInterval:
            return self._has_overlap_single_interval(other)
        else:
            return other.has_overlap(self, match_strand)

    def _has_overlap_single_interval(self, other: Location) -> bool:
        ObjectValidation.require_object_has_type(other, SingleInterval)
        if len(self) == 0 or len(other) == 0:
            return False
        if other.start <= self.start < other.end:
            return True
        if other.start < self.end <= other.end:
            return True
        if self.start <= other.start < self.end:
            return True
        if self.start < other.end <= self.end:
            return True
        return False

    def reverse(self) -> "SingleInterval":
        return self.reverse_strand()

    def reverse_strand(self) -> "SingleInterval":
        return self.reset_strand(self.strand.reverse())

    def reset_strand(self, new_strand: Strand) -> "SingleInterval":
        new_parent = self.parent.strip_location_info() if self.parent else None
        return SingleInterval(self.start, self.end, new_strand, parent=new_parent)

    def reset_parent(self, new_parent: Parent) -> "SingleInterval":
        parent = new_parent.strip_location_info() if new_parent else None
        return SingleInterval(self.start, self.end, self.strand, parent)

    def shift_position(self, shift: int) -> "SingleInterval":
        return SingleInterval(self.start + shift, self.end + shift, self.strand, self.parent)

    def distance_to(self, other: Location, distance_type: DistanceType = DistanceType.INNER) -> int:
        if distance_type == DistanceType.STARTS:
            return abs(self.start - other.start)
        if distance_type == DistanceType.ENDS:
            return abs(self.end - other.end)
        if type(other) is SingleInterval:
            return self._distance_to_single_interval(other, distance_type)
        else:
            return other.distance_to(self, distance_type)

    def _distance_to_single_interval(self, other: Location, distance_type: DistanceType) -> int:
        ObjectValidation.require_object_has_type(other, SingleInterval)
        distances = [abs(self.start - other.end), abs(self.end - other.start)]
        if distance_type == DistanceType.INNER:
            if self.has_overlap(other):
                return 0
            return min(distances)
        if distance_type == DistanceType.OUTER:
            return max(distances)
        raise NotImplementedError(f"Distance type not implemented: {distance_type.value}")

    def intersection(self, other: Location, match_strand: bool = True) -> Location:
        if not self.has_overlap(other, match_strand=match_strand):
            return EmptyLocation()
        if type(other) is SingleInterval:
            return self._intersection_single_interval(other)
        intersect_other_strand = other.intersection(self, match_strand=match_strand)
        return intersect_other_strand.reset_strand(self.strand)

    def _intersection_single_interval(self, other: Location) -> Location:
        ObjectValidation.require_object_has_type(other, SingleInterval)
        new_start = max(self.start, other.start)
        new_end = min(self.end, other.end)
        new_parent = self.parent.strip_location_info() if self.parent else None
        return SingleInterval(new_start, new_end, self.strand, parent=new_parent)

    def union(self, other: Location) -> Location:
        if self.strand != other.strand:
            raise ValueError(f"Strands do not match: {self.strand} != {other.strand}")
        if self.parent:
            ObjectValidation.require_parents_equal_except_location(self.parent, other.parent)
        if type(other) is SingleInterval:
            return self._union_single_interval(other)
        return other.union(self)

    def _union_single_interval(self, other: Location) -> Location:
        ObjectValidation.require_object_has_type(other, SingleInterval)
        new_parent = self.parent.strip_location_info() if self.parent else None
        if len(self) == 0:
            return SingleInterval(other.start, other.end, other.strand, new_parent)
        if len(other) == 0:
            return SingleInterval(self.start, self.end, self.strand, new_parent)
        if self.has_overlap(other):
            return SingleInterval(
                min(self.start, other.start),
                max(self.end, other.end),
                self.strand,
                new_parent,
            )
        else:
            return CompoundInterval.from_single_intervals([self, other])

    def minus(self, other: Location, match_strand: bool = True) -> Location:
        if not self.has_overlap(other, match_strand=match_strand):
            return self
        result_starts = []
        result_ends = []
        curr_result_block_start = self.start
        curr_result_block_end = self.end
        for block in other.blocks:
            if block.contains(self, match_strand=match_strand):
                return EmptyLocation()
            if block.end <= curr_result_block_start:
                continue
            if block.start >= curr_result_block_end:
                break
            if block.start >= curr_result_block_start and block.end <= curr_result_block_end:
                result_starts.append(curr_result_block_start)
                result_ends.append(block.start)
                curr_result_block_start = block.end
            elif block.start < curr_result_block_start and block.end <= curr_result_block_end:
                curr_result_block_start = block.end
            else:
                curr_result_block_end = block.start
                break
        result_starts.append(curr_result_block_start)
        result_ends.append(curr_result_block_end)
        new_parent = self.parent.strip_location_info() if self.parent else None
        return CompoundInterval(result_starts, result_ends, self.strand, parent=new_parent).optimize_blocks()

    def extend_absolute(self, extend_start: int, extend_end: int) -> Location:
        if min(extend_start, extend_end) < 0:
            raise ValueError("Extension distances must be non-negative")
        return SingleInterval(self.start - extend_start, self.end + extend_end, self.strand, self.parent)

    def extend_relative(self, extend_upstream: int, extend_downstream: int) -> Location:
        self.strand.assert_directional()
        if self.strand == Strand.PLUS:
            return self.extend_absolute(extend_upstream, extend_downstream)
        else:
            return self.extend_absolute(extend_downstream, extend_upstream)

    def _location_relative_to(self, other: Location) -> Location:
        intersection = other.intersection(self, match_strand=False)
        rel_pos_1 = other.parent_to_relative_pos(intersection.start)
        rel_pos_2 = other.parent_to_relative_pos(intersection.end - 1)
        rel_start, rel_end = min(rel_pos_1, rel_pos_2), max(rel_pos_1, rel_pos_2) + 1
        rel_strand = self.strand.relative_to(other.strand)
        return SingleInterval(rel_start, rel_end, rel_strand)

    def merge_overlapping(self) -> Location:
        return self

    def to_feature_location(self) -> FeatureLocation:
        """Convert to a BioPython FeatureLocation."""
        return FeatureLocation(self.start, self.end, self.strand.value)

    def to_biopython(self) -> FeatureLocation:
        """Provide a shared function signature with other Locations"""
        return self.to_feature_location()


class CompoundInterval(Location):
    """A location consisting of multiple intervals"""

    def __init__(
        self,
        starts: List[int],
        ends: List[int],
        strand: Strand,
        parent: Optional[ParentInputType] = None,
    ):
        """
        Parameters
        ----------
        starts
            0-based start positions of intervals; need not be sorted but order must match `ends`
        ends
            0-based exclusive end positions of intervals; need not be sorted but order must match `starts`
        strand
            Strand of this Location on parent
        parent
            Parent containing this Location. If parent is not provided including a Sequence,
            some methods requiring the sequence cannot be called. Additionally, some position validation
            cannot be performed. If parent has a location attribute, it is ignored and reset to this location.
        """
        if not len(starts) == len(ends) > 0:
            raise ValueError("Lists of start end end positions must be nonempty and have same length")
        starts_sorted = sorted(starts)
        ends_sorted = sorted(ends)
        parent_obj = make_parent(parent) if parent else None
        self._parent = parent_obj.reset_location(CompoundInterval(starts, ends, strand)) if parent_obj else None
        self._strand = strand
        single_interval_parent = self.parent.strip_location_info() if self.parent else None
        single_intervals_no_parent = [
            SingleInterval(starts_sorted[i], ends_sorted[i], self.strand) for i in range(len(starts_sorted))
        ]
        self._single_intervals = [
            interval.reset_parent(single_interval_parent) for interval in single_intervals_no_parent
        ]
        self._start = self._single_intervals[0].start
        self._end = self._single_intervals[-1].end

    @classmethod
    def from_single_intervals(cls, intervals: List[SingleInterval]) -> "CompoundInterval":
        errors = []
        if not intervals:
            errors.append("List of intervals must be nonempty")
        if len(set([interval.strand for interval in intervals])) > 1:
            errors.append(f"Intervals must all have same strand: {set([interval.strand for interval in intervals])}")
        interval_parents = set(
            [interval.parent.strip_location_info() if interval.parent else None for interval in intervals]
        )
        if len(interval_parents) > 1:
            errors.append(f"Intervals must all have same parent: {set([interval.parent.id for interval in intervals])}")
        if errors:
            raise ValueError("\n".join(errors))
        interval_starts = [interval.start for interval in intervals]
        interval_ends = [interval.end for interval in intervals]
        strand = intervals[0].strand
        new_parent = interval_parents.pop() if interval_parents else None
        return cls(interval_starts, interval_ends, strand, new_parent)

    def __str__(self):
        return ", ".join([str(interval) for interval in self._single_intervals])

    def __eq__(self, other):
        if type(other) is not CompoundInterval:
            return False
        return self.blocks == other.blocks

    def __repr__(self):
        return f"CompoundInterval <{str(self)}>"

    def __len__(self):
        return sum([len(interval) for interval in self._single_intervals])

    @property
    def parent(self) -> Parent:
        return self._parent

    @property
    def strand(self) -> Strand:
        return self._strand

    @property
    def start(self) -> int:
        return self._start

    @property
    def end(self) -> int:
        return self._end

    @property
    def num_blocks(self):
        return len(self._single_intervals)

    @property
    def is_contiguous(self) -> bool:
        return all(
            [self._single_intervals[i + 1].start == self._single_intervals[i].end for i in range(self.num_blocks - 1)]
        )

    @property
    def is_overlapping(self) -> bool:
        """Does this interval have overlaps?"""
        return any(
            [self._single_intervals[i].end > self._single_intervals[i + 1].start for i in range(self.num_blocks - 1)]
        )

    @property
    def is_empty(self) -> bool:
        return self == EmptyLocation()

    @property
    def blocks(self) -> List[Location]:
        return self._single_intervals

    def scan_blocks(self) -> Iterator[Location]:
        self.strand.assert_directional()
        if self.strand == Strand.PLUS:
            for block in self.blocks:
                yield block
        if self.strand == Strand.MINUS:
            for i in range(self.num_blocks):
                yield self.blocks[self.num_blocks - i - 1]

    def extract_sequence(self) -> Sequence:
        self.strand.assert_directional()
        block_seqs = [interval.extract_sequence() for interval in self._single_intervals]
        if self.strand == Strand.PLUS:
            return reduce(lambda seq1, seq2: seq1.append(seq2), block_seqs)
        else:
            return reduce(lambda seq1, seq2: seq1.append(seq2), reversed(block_seqs))

    def parent_to_relative_pos(self, parent_pos: int) -> int:
        rel_pos = 0
        for block in self.scan_blocks():
            try:
                rel_pos += block.parent_to_relative_pos(parent_pos)
                return rel_pos
            except InvalidPositionException:
                rel_pos += len(block)
        raise InvalidPositionException(
            f"Requested parent position ({parent_pos}) lies outside this location ({str(self)})"
        )

    def relative_to_parent_pos(self, relative_pos: int) -> int:
        self.strand.assert_directional()
        if not 0 <= relative_pos < len(self):
            raise InvalidPositionException(
                f"Invalid relative position {relative_pos} for location of length {len(self)}"
            )

        def do_work(rel_pos: int, blocks: List[SingleInterval]):
            plus_strand = self.strand == Strand.PLUS
            block0 = blocks[0] if plus_strand else blocks[-1]
            block0_len = len(block0)
            if rel_pos < block0_len:
                return block0.start + rel_pos if plus_strand else block0.end - 1 - rel_pos
            blocks_tail = blocks[1:] if plus_strand else blocks[:-1]
            return do_work(rel_pos - block0_len, blocks_tail)

        return do_work(relative_pos, self._single_intervals)

    def relative_interval_to_parent_location(
        self, relative_start: int, relative_end: int, relative_strand: Strand
    ) -> Location:
        start_on_parent = self.relative_to_parent_pos(relative_start)
        end_on_parent_inclusive = self.relative_to_parent_pos(relative_end - 1)
        parent_start = min(start_on_parent, end_on_parent_inclusive)
        parent_end = max(start_on_parent, end_on_parent_inclusive) + 1
        intersect_same_strand = self.intersection(
            SingleInterval(
                parent_start,
                parent_end,
                self.strand,
                parent=self.parent.strip_location_info() if self.parent else None,
            )
        )
        return intersect_same_strand.reset_strand(relative_strand.relative_to(self.strand))

    def has_overlap(self, other: Location, match_strand: bool = False) -> bool:
        return any([interval.has_overlap(other, match_strand) for interval in self._single_intervals])

    def optimize_blocks(self) -> Location:
        return self._combine_blocks()._remove_empty_blocks()._to_single_interval_if_one_block()

    def gap_list(self) -> List["Location"]:
        optimized = self.optimize_blocks()
        block_iter = optimized.scan_blocks()
        gaps = []
        block1 = next(block_iter)
        for block2 in block_iter:
            gap_start = min(block1.end, block2.end)
            gap_end = max(block1.start, block2.start)
            gaps.append(SingleInterval(gap_start, gap_end, self.strand, self.parent))
            block1 = block2
        return gaps

    def gaps_location(self) -> "Location":
        gap_list = self.gap_list()
        if gap_list:
            return CompoundInterval.from_single_intervals(gap_list)
        return EmptyLocation()

    def _remove_empty_blocks(self) -> "CompoundInterval":
        return CompoundInterval.from_single_intervals([block for block in self._single_intervals if len(block) > 0])

    def _to_single_interval_if_one_block(self) -> Location:
        return self if self.num_blocks > 1 else self._single_intervals[0]

    def _combine_blocks(self) -> "CompoundInterval":
        new_blocks = []
        curr_block = self._single_intervals[0]
        i = 1
        while i < self.num_blocks:
            next_block = self._single_intervals[i]
            if curr_block.end >= next_block.start:
                new_parent = curr_block.parent.strip_location_info() if curr_block.parent else None
                curr_block = SingleInterval(
                    curr_block.start,
                    next_block.end,
                    curr_block.strand,
                    parent=new_parent,
                )
            else:
                new_blocks.append(curr_block)
                curr_block = next_block
            i += 1
        new_blocks.append(curr_block)
        return CompoundInterval.from_single_intervals(new_blocks)

    def reverse(self) -> "CompoundInterval":
        def reflect_position(relative_pos):
            return self.start + self.end - relative_pos

        def reflect_interval(single_interval: SingleInterval) -> SingleInterval:
            start = reflect_position(single_interval.end)
            end = reflect_position(single_interval.start)
            strand = single_interval.strand.reverse()
            return SingleInterval(start, end, strand)

        new_parent = self.parent.strip_location_info() if self.parent else None
        return CompoundInterval.from_single_intervals(
            [reflect_interval(single_interval) for single_interval in self._single_intervals]
        ).reset_parent(new_parent)

    def reverse_strand(self) -> "CompoundInterval":
        return CompoundInterval.from_single_intervals(
            [interval.reverse_strand() for interval in self._single_intervals]
        )

    def reset_strand(self, new_strand: Strand) -> Location:
        return CompoundInterval.from_single_intervals(
            [interval.reset_strand(new_strand) for interval in self._single_intervals]
        )

    def reset_parent(self, new_parent: Parent) -> "CompoundInterval":
        return CompoundInterval.from_single_intervals(
            [interval.reset_parent(new_parent) for interval in self._single_intervals]
        )

    def shift_position(self, shift: int) -> Location:
        return CompoundInterval.from_single_intervals(
            [interval.shift_position(shift) for interval in self._single_intervals]
        )

    def distance_to(self, other: Location, distance_type: DistanceType = DistanceType.INNER) -> int:
        if distance_type == DistanceType.STARTS:
            return abs(self.start - other.start)
        elif distance_type == DistanceType.ENDS:
            return abs(self.end - other.end)
        elif distance_type == DistanceType.OUTER:
            return max(abs(self.start - other.end), abs(self.end - other.start))
        elif distance_type == DistanceType.INNER:
            distance = None
            for interval in self._single_intervals:
                interval_distance = other.distance_to(interval)
                if distance is None or interval_distance < distance:
                    distance = interval_distance
                    if distance == 0:
                        return 0
            return distance
        else:
            raise NotImplementedError(f"Unknown distance type {distance_type.value}")

    def intersection(self, other: Location, match_strand: bool = True) -> Location:
        if not self.has_overlap(other, match_strand=match_strand):
            return EmptyLocation()
        if type(other) is SingleInterval:
            return self._intersection_single_interval(other, match_strand=match_strand)
        if type(other) is CompoundInterval:
            return self._intersection_compound_interval(other, match_strand=match_strand)
        raise UnsupportedOperationException(f"Not implemented for type {type(other)}")

    def _intersection_single_interval(self, other: Location, match_strand: bool) -> Location:
        ObjectValidation.require_object_has_type(other, SingleInterval)
        interval_intersections = []
        for single_interval in self._single_intervals:
            if single_interval.has_overlap(other, match_strand=match_strand):
                interval_intersections.append(single_interval.intersection(other, match_strand=match_strand))
        return CompoundInterval.from_single_intervals(interval_intersections).optimize_blocks()

    def _intersection_compound_interval(self, other: Location, match_strand: bool) -> Location:
        ObjectValidation.require_object_has_type(other, CompoundInterval)
        interval_intersections = []
        for self_single_interval in self._single_intervals:
            if self_single_interval.has_overlap(other, match_strand=match_strand):
                for other_single_interval in other.blocks:
                    if self_single_interval.has_overlap(other_single_interval, match_strand=match_strand):
                        interval_intersections.append(
                            self_single_interval.intersection(other_single_interval, match_strand=match_strand)
                        )
        return CompoundInterval.from_single_intervals(interval_intersections).optimize_blocks()

    def union(self, other: Location) -> Location:
        if self.strand != other.strand:
            raise ValueError(f"Strands do not match: {self.strand} != {other.strand}")
        if self.parent:
            ObjectValidation.require_parents_equal_except_location(self.parent, other.parent)
        if type(other) is SingleInterval:
            return self._union_single_interval(other)
        if type(other) is CompoundInterval:
            return self._union_compound_interval(other)
        raise UnsupportedOperationException(f"Not implemented for type {type(other)}")

    def _union_single_interval(self, other: Location):
        ObjectValidation.require_object_has_type(other, SingleInterval)
        overlapping_blocks = []
        non_overlapping_blocks = []
        for block in self._single_intervals:
            if block.has_overlap(other):
                overlapping_blocks.append(block)
            else:
                non_overlapping_blocks.append(block)
        if overlapping_blocks:
            start_all_overlappers = min(other.start, min([block.start for block in overlapping_blocks]))
            end_all_overlappers = max(other.end, max([block.end for block in overlapping_blocks]))
            parent_overlappers = (
                overlapping_blocks[0].parent.strip_location_info() if overlapping_blocks[0].parent else None
            )
            interval_all_overlappers = SingleInterval(
                start_all_overlappers,
                end_all_overlappers,
                self.strand,
                parent=parent_overlappers,
            )
            return CompoundInterval.from_single_intervals(
                non_overlapping_blocks + [interval_all_overlappers]
            ).optimize_blocks()
        return CompoundInterval.from_single_intervals(non_overlapping_blocks + [other]).optimize_blocks()

    @staticmethod
    def _merge_compound_blocks(blocks: List[SingleInterval]) -> Location:
        return reduce(lambda left, right: left.union(right), blocks)

    def _union_compound_interval(self, other: Location):
        ObjectValidation.require_object_has_type(other, CompoundInterval)
        blocks = sorted(self.blocks + other.blocks)
        return self._merge_compound_blocks(blocks)

    def minus(self, other: Location, match_strand: bool = True) -> Location:
        if not self.has_overlap(other, match_strand=match_strand):
            return self.optimize_blocks()
        result_blocks = []
        for block in self.blocks:
            block_minus_other = block.minus(other, match_strand=match_strand)
            if block_minus_other is not EmptyLocation():
                result_blocks.extend(block_minus_other.blocks)
        if not result_blocks:
            return EmptyLocation()
        if len(result_blocks) == 1:
            return result_blocks[0]
        return CompoundInterval.from_single_intervals(result_blocks).optimize_blocks()

    def extend_absolute(self, extend_start: int, extend_end: int) -> Location:
        if min(extend_start, extend_end) < 0:
            raise ValueError("Extension distances must be non-negative")
        extended = self
        if extend_start > 0:
            left_flank = SingleInterval(self.start - extend_start, self.start, self.strand, self.parent)
            extended = self.union(left_flank)
        if extend_end > 0:
            right_flank = SingleInterval(self.end, self.end + extend_end, self.strand, self.parent)
            extended = extended.union(right_flank)
        return extended.optimize_blocks()

    def extend_relative(self, extend_upstream: int, extend_downstream: int) -> Location:
        self.strand.assert_directional()
        if self.strand == Strand.PLUS:
            return self.extend_absolute(extend_upstream, extend_downstream)
        else:
            return self.extend_absolute(extend_downstream, extend_upstream)

    def _location_relative_to(self, other: Location) -> Location:
        rel_loc_each_single_interval = [
            other.parent_to_relative_location(block) for block in self.blocks if other.has_overlap(block)
        ]
        return reduce(
            lambda location1, location2: location1.union(location2),
            rel_loc_each_single_interval,
        ).optimize_blocks()

    def merge_overlapping(self) -> Location:
        """If this compound interval is overlapping, merge the overlaps"""
        if not self.is_overlapping:
            return self
        return self._merge_compound_blocks(self.blocks)

    def to_compound_location(self) -> CompoundLocation:
        """Convert to a BioPython CompoundLocation, or FeatureLocation if this is a 1-block interval"""
        blocks = [b.to_biopython() for b in self.blocks]
        if len(blocks) == 1:
            return blocks[0]
        return CompoundLocation(blocks)

    def to_biopython(self) -> CompoundLocation:
        """Provide a shared function signature with other Locations"""
        return self.to_compound_location()


class _EmptyLocation(Location):
    """Singleton object representing an empty location"""

    def __len__(self):
        return 0

    def __str__(self):
        return "EmptyLocation"

    def __eq__(self, other):
        return type(other) is _EmptyLocation

    def __repr__(self):
        return "EmptyLocation"

    @property
    def parent(self) -> Parent:
        return None

    @property
    def strand(self) -> Strand:
        raise EmptyLocationException

    @property
    def start(self) -> int:
        raise EmptyLocationException

    @property
    def end(self) -> int:
        raise EmptyLocationException

    @property
    def is_contiguous(self) -> bool:
        raise EmptyLocationException

    @property
    def is_empty(self) -> bool:
        return True

    @property
    def blocks(self) -> List[Location]:
        return []

    def scan_blocks(self) -> None:
        return

    @property
    def num_blocks(self) -> int:
        return 0

    @property
    def is_overlapping(self) -> bool:
        """EmptyLocation is by definition always non-overlapping"""
        return False

    def optimize_blocks(self) -> Location:
        return self

    def gap_list(self) -> List["Location"]:
        return []

    def gaps_location(self) -> "Location":
        return self

    def extract_sequence(self) -> Sequence:
        raise EmptyLocationException

    def parent_to_relative_pos(self, parent_pos: int) -> int:
        raise EmptyLocationException

    def relative_to_parent_pos(self, relative_pos: int) -> int:
        raise EmptyLocationException

    def parent_to_relative_location(self, parent_location) -> Location:
        raise EmptyLocationException

    def relative_interval_to_parent_location(
        self, relative_start: int, relative_end: int, relative_strand: Strand
    ) -> Location:
        raise EmptyLocationException

    def has_overlap(self, other: Location, match_strand: bool = False) -> bool:
        return False

    def reverse(self) -> Location:
        return self

    def reverse_strand(self) -> Location:
        return self

    def reset_strand(self, new_strand: Strand) -> Location:
        raise EmptyLocationException

    def reset_parent(self, new_parent: Parent) -> Location:
        raise EmptyLocationException

    def shift_position(self, shift: int) -> Location:
        raise EmptyLocationException

    def location_relative_to(self, other: Location) -> Location:
        return self

    def _location_relative_to(self, other: Location) -> Location:
        return self

    def distance_to(self, other: Location, distance_type: DistanceType = DistanceType.INNER) -> int:
        raise EmptyLocationException

    def intersection(self, other: Location, match_strand: bool = True) -> Location:
        return self

    def union(self, other: Location) -> Location:
        raise EmptyLocationException

    def minus(self, other: Location, match_strand: bool = True) -> Location:
        return self

    def extend_absolute(self, extend_start: int, extend_end: int) -> Location:
        raise EmptyLocationException

    def extend_relative(self, extend_upstream: int, extend_downstream: int) -> Location:
        raise EmptyLocationException

    def merge_overlapping(self) -> Location:
        return self

    def to_biopython(self):
        raise EmptyLocationException

    def first_ancestor_of_type(self, sequence_type: str) -> Parent:
        raise EmptyLocationException

    _instance = None


def EmptyLocation():
    """Returns the single EmptyLocation instance"""
    if _EmptyLocation._instance is None:
        _EmptyLocation._instance = _EmptyLocation()
    return _EmptyLocation._instance



import itertools
import re
from pathlib import Path

from setuptools import find_namespace_packages, setup


def get_version() -> str:
    """Return package version as listed in `__version__` in `init.py`.

    Adapted from ``setup.py`` from https://github.com/encode/httpx.

    """
    version = Path("inscripta", "biocantor", "__init__.py").read_text()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", version).group(1)


dependencies = []

extra_dependencies = {
    "libraries": ["biopython"],
    "test": ["black", "flake8", "pytest", "pytest-cov", "pytest-flake8"],
    "docs": [
        "Sphinx",
        "sphinx_rtd_theme",
        "sphinx-automodapi",
        "graphviz",
        "nbconvert",
        "nbsphinx",
        "ipykernel",
        "pandoc",
    ],
    "optional": ["tornado>=5.1"],
}

all_dependencies = list(itertools.chain.from_iterable(extra_dependencies.values()))
extra_dependencies["all"] = all_dependencies

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="BioCantor",
    version=get_version(),
    description="Flexible feature arithmetic, seamlessly integrated with nested coordinate systems.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Inscripta, Inc.",
    url="https://github.com/InscriptaLabs/BioCantor",
    test_suite="pytest",
    packages=find_namespace_packages(include=["inscripta.*"]),
    include_package_data=True,
    tests_require=extra_dependencies["test"],
    extras_require=extra_dependencies,
    install_requires=dependencies,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from abc import ABC, abstractmethod
from typing import Iterator, List, Union

from Bio.SeqFeature import FeatureLocation, CompoundLocation

from inscripta.biocantor.exc import NoSuchAncestorException
from inscripta.biocantor.location.distance import DistanceType
from inscripta.biocantor.location.strand import Strand
from inscripta.biocantor.parent import Parent
from inscripta.biocantor.sequence import Sequence
from inscripta.biocantor.util.object_validation import ObjectValidation


class Location(ABC):
    """Abstract location with respect to a coordinate system"""

    @abstractmethod
    def __len__(self):
        """Returns the length (number of positions) of this Location. For subclasses representing discontiguous
        locations, regions between blocks are not considered."""

    @abstractmethod
    def __str__(self):
        """Returns a human readable string representation of this Location"""

    @abstractmethod
    def __eq__(self, other):
        """Returns True iff this Location is equal to other object"""

    @abstractmethod
    def __repr__(self):
        """Returns the 'official' string representation of this Location"""

    @property
    @abstractmethod
    def parent(self) -> Parent:
        """Returns the parent of this Location"""

    @property
    def parent_id(self) -> str:
        """Returns the parent ID"""
        return self.parent.id if self.parent else None

    @property
    def parent_type(self) -> str:
        """Returns the sequence type of the parent"""
        return self.parent.sequence_type if self.parent else None

    @property
    @abstractmethod
    def strand(self) -> Strand:
        """Returns the strand of this Location with respect to its parent"""

    @property
    @abstractmethod
    def start(self) -> int:
        """Returns the 0-based start position of this Location on its parent"""

    @property
    @abstractmethod
    def end(self) -> int:
        """Returns the 0-based exclusive end position of this Location on its parent"""

    @property
    @abstractmethod
    def is_contiguous(self) -> bool:
        """Returns True iff this Location is fully contiguous within its parent"""

    @property
    @abstractmethod
    def is_empty(self) -> bool:
        """Returns True iff this Location is empty"""

    @property
    @abstractmethod
    def blocks(self) -> List["Location"]:
        """Returns list of contiguous blocks comprising this Location"""

    @property
    @abstractmethod
    def is_overlapping(self) -> bool:
        """Returns True if this interval contains overlaps; always False for SingleInterval"""

    @abstractmethod
    def scan_blocks(self) -> Iterator["Location"]:
        """Returns an iterator over blocks in order relative to strand of this Location"""

    @property
    @abstractmethod
    def num_blocks(self) -> int:
        """Returns number of contiguous blocks comprising this Location"""

    @abstractmethod
    def optimize_blocks(self) -> "Location":
        """Returns a new Location covering the same positions but with blocks optimized.
        For example, empty blocks may be removed or adjacent blocks may be combined if applicable."""

    @abstractmethod
    def gap_list(self) -> List["Location"]:
        """Returns list of contiguous regions comprising the space between blocks of this Location. List is
        ordered relative to strand of this Location."""

    @abstractmethod
    def gaps_location(self) -> "Location":
        """Returns a Location representing the space between blocks of this Location."""

    @abstractmethod
    def extract_sequence(self) -> Sequence:
        """Extracts the sequence of this Location from the parent.
        Concrete implementations should raise ValueError if no parent exists."""

    @abstractmethod
    def parent_to_relative_pos(self, parent_pos: int) -> int:
        """Converts a position on the parent to a position relative to this Location.
        Concrete implementations should raise ValueError if the given position does not overlap this Location."""

    @abstractmethod
    def relative_to_parent_pos(self, relative_pos: int) -> int:
        """Converts a position relative to this Location to a position on the parent"""

    def parent_to_relative_location(self, parent_location: "Location") -> "Location":
        """Converts a Location on the parent to a Location relative to this Location.

        Parameters
        ----------
        parent_location
            Location with the same parent as this Location. Both parents can be None.

        Returns
        -------
        New Location relative to this Location.
        """
        return parent_location.location_relative_to(self)

    def location_relative_to(self, other: "Location") -> "Location":
        """Converts this Location to a Location relative to another Location. The Locations must overlap.
        The returned value represents the relative location of the overlap within the other Location."""
        if other.is_empty:
            return other
        if self.parent or other.parent:
            if not self.parent and other.parent:
                raise ValueError(
                    "Parents must be both null or both non-null:\n{}\n  !=\n{}".format(self.parent, other.parent)
                )
            ObjectValidation.require_parents_equal_except_location(self.parent, other.parent)
        ObjectValidation.require_locations_overlap(self, other)
        return self._location_relative_to(other)

    @abstractmethod
    def _location_relative_to(self, other: "Location") -> "Location":
        raise NotImplementedError

    @abstractmethod
    def relative_interval_to_parent_location(
        self, relative_start: int, relative_end: int, relative_strand: Strand
    ) -> "Location":
        """Converts an interval relative to this Location to a Location on the parent

        Parameters
        ----------
        relative_start
            0-based start position of interval relative to this Location
        relative_end
            0-based exclusive end position of interval relative to this Location
        relative_strand
            Strand of interval relative to the strand of this Location. If the strand of interval is on the SAME
            strand as the strand of this location, relative_strand is PLUS. If the strand interval is on the OPPOSITE
            strand, relative_strand is MINUS.

        Returns
        -------
        New Location on the parent with the parent as parent
        """

    def scan_windows(self, window_size: int, step_size: int, start_pos: int = 0) -> Iterator["Location"]:
        """Returns an iterator over fixed size windows within this Location. Windows represent sub-regions
        of this Location and are with respect to the same parent as this Location. The final window returned
        is the last one that fits completely within this Location. Returned windows are in order according to
        relative position within this Location; i.e., corresponding to the strand of this Location.

        Parameters
        ----------
        window_size
        step_size
        start_pos
            0-based relative start position of first window relative to this Location
        """
        if not 0 <= start_pos < len(self):
            raise ValueError("Start position ({}) must be within location length ({})".format(start_pos, len(self)))
        if min(window_size, step_size) < 1:
            raise ValueError("Window size and step size must both be positive: {}, {}".format(window_size, step_size))
        if window_size > len(self):
            raise ValueError("Window size ({}) must be <= location length ({})".format(window_size, len(self)))
        if start_pos + window_size > len(self):
            raise ValueError(
                "Start position ({}) + window size ({}) must be <= location length ({})".format(
                    start_pos, window_size, len(self)
                )
            )
        self.strand.assert_directional()
        for curr_start in range(start_pos, len(self) - window_size + 1, step_size):
            yield self.relative_interval_to_parent_location(curr_start, curr_start + window_size, Strand.PLUS)

    @abstractmethod
    def has_overlap(self, other: "Location", match_strand: bool = False) -> bool:
        """Returns True iff this Location shares at least one position with the given Location.
        For subclasses representing discontiguous locations, regions between blocks are not considered.

        Parameters
        ----------
        other
            Other Location
        match_strand
            If set to True, automatically return False if given interval Strand does not match this Location's Strand

        Returns
        -------
        True if there is any overlap, False otherwise
        """

    @abstractmethod
    def reverse(self) -> "Location":
        """Returns a new Location corresponding to this Location with the same start and stop, with
        strand and structure reversed"""

    @abstractmethod
    def reverse_strand(self) -> "Location":
        """Returns a new Location corresponding to this Location with the strand reversed"""

    @abstractmethod
    def reset_strand(self, new_strand: Strand) -> "Location":
        """Returns a new Location corresponding to this Location with the given strand"""

    @abstractmethod
    def reset_parent(self, new_parent: Parent) -> "Location":
        """Returns a new Location corresponding to this Location with positions unchanged and pointing
        to a new parent"""

    @abstractmethod
    def shift_position(self, shift: int) -> "Location":
        """Returns a new Location corresponding to this location shifted by the given distance"""

    @abstractmethod
    def distance_to(self, other: "Location", distance_type: DistanceType = DistanceType.INNER) -> int:
        """Returns the distance from this location to another location with the same parent.
        Return value is a non-negative integer and implementations must be commutative.

        Parameters
        ----------
        other
            Other location with same parent as this location
        distance_type
            Distance type
        """

    @abstractmethod
    def merge_overlapping(self) -> "Location":
        """Merges overlapping windows"""

    @abstractmethod
    def to_biopython(self) -> Union[FeatureLocation, CompoundLocation]:
        """Returns a BioPython interval type; since they do not have a shared base class, we need a union"""
        pass

    def first_ancestor_of_type(self, sequence_type: str) -> Parent:
        """Returns the Parent object representing the closest ancestor (parent, parent of parent, etc.)
        of this location which has the given sequence type. Raises NoSuchAncestorException if no ancestor with
        the given type exists."""
        if not self.parent:
            raise NoSuchAncestorException("Location has no parent")
        return self.parent.first_ancestor_of_type(sequence_type, include_self=True)

    def has_ancestor_of_type(self, sequence_type: str) -> bool:
        """Returns True if some ancestor (parent, parent of parent, etc.) of of this location has the given sequence
        type, or False otherwise."""
        if not self.parent:
            return False
        return self.parent.has_ancestor_of_type(sequence_type, include_self=True)

    def lift_over_to_first_ancestor_of_type(self, sequence_type: str) -> "Location":
        """Returns a new Location representing the liftover of this Location to its closest ancestor sequence (parent,
        parent of parent, etc.) which has the given sequence type. If the immediate parent has the given type,
        returns this Location. Raises NoSuchAncestorException if no ancestor with the given type exists."""
        try:
            self.first_ancestor_of_type(sequence_type)
        except NoSuchAncestorException:
            raise NoSuchAncestorException("Location has no ancestor of type {}".format(sequence_type))
        if self.parent_type == sequence_type:
            return self
        lifted_to_grandparent = self.parent.lift_child_location_to_parent()
        return lifted_to_grandparent.lift_over_to_first_ancestor_of_type(sequence_type)

    def has_ancestor_sequence(self, sequence: Sequence) -> bool:
        """Returns True iff this Location has some ancestor (parent, parent of parent, etc.) whose sequence
        attribute is equal to the given sequence"""
        if not self.parent:
            return False
        return self.parent.has_ancestor_sequence(sequence, include_self=True)

    def lift_over_to_sequence(self, sequence: Sequence) -> "Location":
        """Returns a new Location representing the liftover of this Location to the given sequence. The given
        sequence must be equal to the sequence attribute of some Parent in the ancestor hierarchy of this
        Location; otherwise, raises NoSuchAncestorException."""
        if not self.is_contiguous:
            raise ValueError("Location must be contiguous")
        if not self.has_ancestor_sequence(sequence):
            raise NoSuchAncestorException(
                "\nLocation:\n{}\nDoes not have ancestor:\n{}".format(str(self), sequence.summary())
            )
        if self.parent.sequence and self.parent.sequence == sequence:
            return self
        lifted_to_grandparent = self.parent.lift_child_location_to_parent()
        return lifted_to_grandparent.lift_over_to_sequence(sequence)

    @abstractmethod
    def intersection(self, other: "Location", match_strand: bool = True) -> "Location":
        """Returns a new Location representing the intersection of this Location with the other Location.
        Returned Location, if nonempty, has the same Strand as this Location. This operation is commutative
        if match_strand is True.

        Parameters
        ----------
        other
            Other location
        match_strand
            If set to True, automatically return EmptyLocation() if other Location has a different Strand than
            this Location
        """

    @abstractmethod
    def union(self, other: "Location") -> "Location":
        """Returns a new Location representing the union of this Location with the other Location. This operation
        is commutative. Raises ValueError if locations cannot be combined."""

    @abstractmethod
    def minus(self, other: "Location", match_strand: bool = True) -> "Location":
        """Returns a new Location representing this Location minus its intersection with the other Location.
        Returned Location has the same Strand as this Location. If there is no intersection, returns this Location.
        This operation is not commutative.

        Parameters
        ----------
        other
            Other location
        match_strand
            If set to True, automatically return this Location if other Location has a different Strand than this
            Location
        """

    @abstractmethod
    def extend_absolute(self, extend_start: int, extend_end: int) -> "Location":
        """Returns a new Location representing this Location with start and end positions extended by the given
        values, ignoring Strand. Returned Location has same Strand as this Location.

        Parameters
        ----------
        extend_start
            Non-negative integer: amount to extend start
        extend_end
            Non-negative integer: amount to extend end
        """

    @abstractmethod
    def extend_relative(self, extend_upstream: int, extend_downstream: int) -> "Location":
        """Returns a new Location extended upstream and downstream relative to this Location's Strand.

        Parameters
        ----------
        extend_upstream
            Non-negative integer: amount to extend upstream relative to Strand
        extend_downstream
            Non-negative integer: amount to extend downstream relative to Strand
        """

    def contains(self, other: "Location", match_strand: bool = False):
        """Returns True iff this location contains the other"""
        if not self.has_overlap(other, match_strand):
            return False
        other_to_compare = other.reset_parent(None)
        return len(self.reset_parent(None).intersection(other_to_compare, match_strand=match_strand)) == len(
            other_to_compare
        )
